/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2016                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/
/*! \file trimesh_base.cpp
\ingroup code_sample

\brief the minimal example of using the lib

This file contain a minimal example of the library, showing how to load a mesh and how to compute per vertex normals on it.

*/

#include<vcg/complex/complex.h>
#include<vcg/complex/algorithms/clean.h>

#include<wrap/io_trimesh/import_off.h>
#include<wrap/io_trimesh/export_ply.h>

class MyVertex; class MyEdge; class MyFace; class MyRealEdge;
struct MyUsedTypes : public vcg::UsedTypes<vcg::Use<MyVertex>   ::AsVertexType,
                                           vcg::Use<MyEdge>     ::AsEdgeType,
                                           vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex< MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::Color4b, vcg::vertex::BitFlags  >{};
class MyFace    : public vcg::Face<   MyUsedTypes, vcg::face::FFAdj, vcg::face::Normal3f, vcg::face::Color4b,  vcg::face::Qualityf, vcg::face::VertexRef, vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge< MyUsedTypes,vcg::edge::VertexRef, vcg::edge::BitFlags> {};

class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> > {};
class MyEdgeMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>,  std::vector<MyEdge>, std::vector<MyFace> > {};

using namespace vcg;

void ExtrudeFace(MyMesh &m, int faceId, float offset)
{
    m.face[faceId].C()=Color4b::Red;
    MyMesh::VertexPointer nvp[3];
    MyMesh::VertexPointer vp[3];
    
    auto vi = &* tri::Allocator<MyMesh>::AddVertices(m,3);
    for(int i=0;i<3;++i)
    {
        nvp[i] = &*vi++;
        Point3f np = m.face[faceId].V(i)->P() + offset*m.face[faceId].V(i)->N();
        nvp[i]->P() =np;
        vp[i] = m.face[faceId].V(i);
    }
    printf("Mesh has %i %i\n",m.vn,m.fn);
    
    tri::Allocator<MyMesh>::AddFace(m,vp[0], vp[1],nvp[1]);
    tri::Allocator<MyMesh>::AddFace(m,vp[0],nvp[1],nvp[0]);
    
    tri::Allocator<MyMesh>::AddFace(m,vp[1], vp[2],nvp[2]);
    tri::Allocator<MyMesh>::AddFace(m,vp[1],nvp[2],nvp[1]);
    
    tri::Allocator<MyMesh>::AddFace(m,vp[2], vp[0],nvp[0]);
    tri::Allocator<MyMesh>::AddFace(m,vp[2],nvp[0],nvp[2]);
    
    for(int i=0;i<3;++i)
        m.face[faceId].V(i)=nvp[i];
}

void RefineMesh(MyMesh &m, MyMesh &refinedMesh)
{
    for(MyMesh::FaceIterator fi=m.face.begin();fi!=m.face.end();++fi)
    {
        std::vector<size_t> newFaceIndVec;
        for(int i=0;i<3;++i)
        {
            newFaceIndVec.push_back(refinedMesh.VN());
            tri::Allocator<MyMesh>::AddVertex(refinedMesh, fi->P0(i));
            
            // Point3f newp = fi->V(0)->P()*0.5f + fi->V(0)->P()*0.25f + fi->V(0)->P()*0.25f ;
            Point3f newp = fi->P0(i)*0.5f + fi->P1(i)*0.25f + fi->P2(i)*0.25f ;
            Point3f b= vcg::Barycenter(*fi);
            newp = b + ((fi->P(i)-b) *0.5f);
            
            newFaceIndVec.push_back(refinedMesh.VN());
            tri::Allocator<MyMesh>::AddVertex(refinedMesh, newp);
            
            newp= fi->P0(i)*2.0f/3.0f + fi->P1(i)*1.0f/3.0f;
            newFaceIndVec.push_back(refinedMesh.VN());
            tri::Allocator<MyMesh>::AddVertex(refinedMesh, newp);
            newp= fi->P0(i)*1.0f/3.0f + fi->P1(i)*2.0f/3.0f;
            newFaceIndVec.push_back(refinedMesh.VN());
            tri::Allocator<MyMesh>::AddVertex(refinedMesh, newp);
        }
        
        assert(newFaceIndVec.size()==12);
        
        tri::Allocator<MyMesh>::AddFace(refinedMesh,newFaceIndVec[1],newFaceIndVec[5],newFaceIndVec[9]);
        
        for(int i=0;i<3;++i)
        {
            tri::Allocator<MyMesh>::AddFace(refinedMesh,newFaceIndVec[0+i*4],newFaceIndVec[2+i*4],newFaceIndVec[1+i*4]);
            tri::Allocator<MyMesh>::AddFace(refinedMesh,newFaceIndVec[2+i*4],newFaceIndVec[3+i*4],newFaceIndVec[1+i*4]);
            tri::Allocator<MyMesh>::AddFace(refinedMesh,newFaceIndVec[1+i*4],newFaceIndVec[3+i*4],newFaceIndVec[(5+i*4)%12]);
            tri::Allocator<MyMesh>::AddFace(refinedMesh,newFaceIndVec[(5+i*4)%12],newFaceIndVec[3+i*4],newFaceIndVec[(4+i*4)%12]);          
        }
    }
    
    tri::Clean<MyMesh>::RemoveDuplicateVertex(refinedMesh);
    tri::UpdateNormal<MyMesh>::PerVertexNormalized(refinedMesh);
    tri::UpdateNormal<MyMesh>::PerFaceNormalized(refinedMesh);    
}

int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.off>\n");
    return -1;
  }
  /*!
    */
  MyMesh m;
  MyEdgeMesh em;

  if(tri::io::ImporterOFF<MyMesh>::Open(m,argv[1])!=tri::io::ImporterOFF<MyMesh>::NoError)
  {
    printf("Error reading file  %s\n",argv[1]);
    exit(0);
  }

  tri::RequirePerVertexNormal(m);
  tri::UpdateNormal<MyMesh>::PerVertexNormalized(m);
  tri::UpdateNormal<MyMesh>::PerFaceNormalized(m);
  
  tri::UpdateBounding<MyMesh>::Box(m);
  float offset = m.bbox.Diag()/200.0f;
  
  printf("Mesh has %i %i\n",m.vn,m.fn);
   
  
  tri::UpdateFlags<MyMesh>::FaceClearV(m);
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  int connectedComponentCnt=0;
  int faceCnt=0;
  
  MyMesh refM;
  RefineMesh(m,refM);

  printf("Start Visiting %i\n",connectedComponentCnt); fflush(stdout);
 
  
  for(MyMesh::FaceIterator fi=m.face.begin();fi!=m.face.end();++fi)
  {
      if(!fi->IsV())
      {
          printf("Start Visiting %i\n",connectedComponentCnt); fflush(stdout);
          connectedComponentCnt++;
          std::vector<MyMesh::FacePointer> faceStack;
          faceStack.push_back(&*fi);
          fi->SetV();
          while(!faceStack.empty())
          {
              std::swap(faceStack.back(),faceStack[rand()%faceStack.size()]);
              MyMesh::FacePointer fp=faceStack.back();
              faceStack.pop_back();
              faceCnt++;
              fp->C().SetColorRamp(0,m.FN(),faceCnt);
              fp->Q()=faceCnt;
              int basefaceInd = tri::Index(m,fp);
              for(int i=0;i<3;++i)
              {
                  ExtrudeFace(refM,1+basefaceInd*13+1*i,offset);
                  ExtrudeFace(refM,1+basefaceInd*13+4*i,offset);
              }
              
              for(int i=0;i<3;++i)
              {
                  if(!fp->FFp(i)->IsV())
                  {
                      faceStack.push_back(fp->FFp(i));
                      fp->FFp(i)->SetV();
                      tri::Allocator<MyEdgeMesh>::AddEdge(em, vcg::Barycenter(*fp)+fp->N()*offset, vcg::Barycenter(*(fp->FFp(i)))+fp->FFp(i)->N()*offset);                          
                  }
              }
          }            
      }
  }
  
  tri::io::ExporterPLY<MyMesh>::Save(m,"Test.ply",tri::io::Mask::IOM_FACECOLOR+tri::io::Mask::IOM_VERTCOLOR);
  

  tri::Clean<MyEdgeMesh>::RemoveDuplicateVertex(em);
  tri::io::ExporterPLY<MyEdgeMesh>::Save(em,"Edge.ply",tri::io::Mask::IOM_EDGEINDEX);
  
  tri::io::ExporterPLY<MyMesh>::Save(refM,"refined.ply");
  
  return 0;
}
